import logging
import time
from datetime import timedelta
from typing import NamedTuple
from IPython.display import display, clear_output
import ipywidgets as widgets

import clease
from clease.settings import CEBulk, CECrystal
from clease.corr_func import CorrFunction

from clease_gui import register_logger, utils
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar
from . import (
    ConcentrationDashboard,
    StructureSettingsDashboard,
    ClusterDashboard,
    CESettingsDashboard,
)

__all__ = ["SettingsMakerDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class _SubDashboard(NamedTuple):
    dashboard: BaseDashboard  # Instance of the Dashboard
    tab_name: str

    def make_output(self) -> widgets.Output:
        """Make a new output object, and draw the dashboard on it."""
        out = widgets.Output()
        with out:
            clear_output()
            self.dashboard.display()
        return out


def _make_subtab(*sub_dashboards: _SubDashboard) -> widgets.Tab:
    """Construct the tabs in the sub-dashboards, and initialize their displays"""
    children = [sub.make_output() for sub in sub_dashboards]
    tab = widgets.Tab(children=children)
    for ii, sub in enumerate(sub_dashboards):
        tab.set_title(ii, sub.tab_name)
    return tab


class SettingsMakerDashboard(BaseDashboard):
    """Have the concentration and settings dashboard under 1 tab"""

    def initialize(self):
        self.tab = None

        self.make_settings_btn = utils.make_clickable_button(
            self._on_make_settings_click, description="Make settings"
        )

        self.delete_settings_btn = utils.make_clickable_button(
            self._on_delete_settings_click,
            description="Delete settings",
            button_style="danger",
        )

        self.save_settings_fname = widgets.Text(description="Filename:", value="settings.json")
        self.save_settings_btn = utils.make_clickable_button(
            self._on_save_settings_click, description="Save settings"
        )

        self.save_settings_box = widgets.HBox(
            children=[
                self.save_settings_btn,
                self.save_settings_fname,
            ]
        )

        self.load_settings_fname = widgets.Text(description="Filename:", value="settings.json")
        self.load_settings_btn = utils.make_clickable_button(
            self._on_load_settings_click, description="Load settings"
        )

        self.load_settings_box = widgets.HBox(
            children=[
                self.load_settings_btn,
                self.load_settings_fname,
            ]
        )

        self._conc_dashboard = ConcentrationDashboard(self.app_data)
        self._crystal_settings_dashboard = StructureSettingsDashboard(self.app_data)
        self._ce_settings_dashboard = CESettingsDashboard(self.app_data)
        self._cluster_dashboard = ClusterDashboard(self.app_data)

        self.tab = _make_subtab(
            _SubDashboard(self._conc_dashboard, "Concentration"),
            _SubDashboard(self._crystal_settings_dashboard, "Crystal Settings"),
            _SubDashboard(self._ce_settings_dashboard, "CE Settings"),
            _SubDashboard(self._cluster_dashboard, "Clusters"),
        )

        self.reconfig_db_button = utils.make_clickable_button(
            self._on_reconfig_click, description="Reconfigure DB"
        )
        self.reconfigure_progress = widgets.IntProgress()
        self.reconfigure_progress_label = widgets.Label()

    @update_statusbar
    @utils.disable_cls_widget("reconfig_db_button")
    def _on_reconfig_click(self, b):
        with self.event_context(logger=logger):
            self.reconfig_db()

    def reconfig_db(self):
        settings = self.settings
        if settings is None:
            raise ValueError("Settings has not been created yet, cannot reconfigure")
        # TODO: Account for CWD?
        logger.info("Reconfiguring database: %s", settings.db_name)

        cf = CorrFunction(settings)
        self.reconfigure_progress.value = 0
        self.reconfigure_progress_label.value = "Starting..."
        t_start = time.perf_counter()
        for row_id, count, total in cf.iter_reconfigure_db_entries():
            dt = timedelta(seconds=int(time.perf_counter() - t_start))
            msg = f"{count}/{total}, ID: {row_id}. Runtime: {dt}"
            self.reconfigure_progress.value = count
            self.reconfigure_progress.max = total
            self.reconfigure_progress_label.value = msg

        dt = timedelta(seconds=time.perf_counter() - t_start)
        logger.info(f"Reconfiguration complete in {dt}.")

    def display(self):
        hbox_btn = widgets.HBox(children=[self.make_settings_btn, self.delete_settings_btn])
        reconfig_widgets = widgets.HBox(
            children=[
                self.reconfig_db_button,
                self.reconfigure_progress,
                self.reconfigure_progress_label,
            ]
        )
        display(
            hbox_btn,
            self.save_settings_box,
            self.load_settings_box,
            reconfig_widgets,
            self.tab,
        )

    def get_concentration(self):
        return self._conc_dashboard.get_concentration()

    @update_statusbar
    @utils.disable_cls_widget("make_settings_btn")
    def _on_make_settings_click(self, b):
        if self.settings is not None:
            logger.error("Cannot make settings, a settings object already exists.")
            return
        logger.info("Making settings...")
        with self.event_context(logger=logger) as cm:
            self.make_settings()
        if not cm.had_error:
            logger.info("Done!")

    @update_statusbar
    @utils.disable_cls_widget("delete_settings_btn")
    def _on_delete_settings_click(self, b):
        try:
            del self.app_data[self.KEYS.SETTINGS]
            logger.info("Removed settings from app data")
        except KeyError:
            logger.info("No settings in app data, cannot delete.")

    def make_settings(self) -> None:
        """Combine the settings in the relevant dashboards, and create the settings"""
        conc = self.get_concentration()
        crystal_kwargs = self._crystal_settings_dashboard.get_settings_kwargs()
        ce_kwargs = self._ce_settings_dashboard.get_settings_kwargs()
        kwargs = dict(**crystal_kwargs, **ce_kwargs)

        settings_type = kwargs.pop("type")
        skew_threshold = kwargs.pop("skew_threshold", None)

        if settings_type == "CEBulk":
            settings = CEBulk(conc, **kwargs)
        elif settings_type == "CECrystal":
            settings = CECrystal(conc, **kwargs)
        elif settings_type == "CESlab":
            # miller = settings_type.pop('miller')
            raise NotImplementedError("Not yet implemneted, yo!")
        else:
            raise ValueError(f"Something is wrong. Got unknown settings_type {settings_type}")

        self.set_settings(settings)
        if skew_threshold is not None:
            self.settings.skew_threshold = skew_threshold

    @property
    def settings(self):
        """Get the settings from the app data"""
        return self.app_data.get(self.KEYS.SETTINGS, None)

    def set_settings(self, settings):
        """Insert settings into the app data"""
        self.app_data[self.KEYS.SETTINGS] = settings

    @property
    def cwd(self):
        return self.app_data[self.KEYS.CWD]

    def get_save_settings_fname(self):
        return self.cwd / self.save_settings_fname.value

    @update_statusbar
    @utils.disable_cls_widget("save_settings_btn")
    def _on_save_settings_click(self, b):
        with self.event_context(logger=logger):
            fname = self.get_save_settings_fname()
            logger.info("Saving settings in file: %s", str(fname))
            self.save_settings(fname)
            logger.info("Save successful.")

    def save_settings(self, fname):
        settings = self.settings
        if settings is None:
            raise RuntimeError("Settings has not been created yet.")
        settings.save(fname)

    def load_settings(self, fname) -> None:
        settings = clease.settings.settings_from_json(fname)
        logger.debug("Loaded settings: %s", settings.todict())
        self.set_settings(settings)
        self.set_settings_widget_states(settings)

    def get_load_settings_fname(self):
        return self.app_data[self.KEYS.CWD] / self.load_settings_fname.value

    @update_statusbar
    @utils.disable_cls_widget("load_settings_btn")
    def _on_load_settings_click(self, b):
        with self.event_context(logger=logger):
            fname = self.get_load_settings_fname()
            logger.info("Loading settings from file: %s", str(fname))
            self.load_settings(fname)
            logger.info("Load successful.")

    def set_settings_widget_states(self, settings):
        """Set the widget states from a new settings object"""
        conc = settings.concentration
        self._conc_dashboard.set_widgets_from_load(conc)
        self._crystal_settings_dashboard.set_widgets_from_load(settings)
        self._ce_settings_dashboard.set_widgets_from_load(settings)



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets

from clease_gui import utils, register_logger
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar

__all__ = ["ClusterDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class ClusterDashboard(BaseDashboard):
    """Dashboard for printing and visualizing the available clusters/figures
    in the active settings object."""

    def initialize(self) -> None:
        # Output for displaying clusters table
        self.clusters_content_out = widgets.Output(
            layout=dict(
                overflow="auto",
                width="100%",
                height="350px",
            )
        )

        self.refresh_table_button = utils.make_clickable_button(
            self._on_refresh_table_click, description="Update Clusters Table"
        )

        self.view_clusters_button = utils.make_clickable_button(
            self._on_view_clusters_click, description="View Clusters"
        )

    def display(self) -> None:
        button_box = widgets.HBox(children=[self.refresh_table_button, self.view_clusters_button])
        display(button_box, self.clusters_content_out)

    @utils.disable_cls_widget("refresh_table_button")
    @update_statusbar
    def _on_refresh_table_click(self, b) -> None:
        """Event handler for when refresh table button is clicked."""
        with self.event_context(logger=logger):
            logger.info("Updating cluster table.")
            self.refresh_clusters_table()

    def refresh_clusters_table(self) -> None:
        """Draw the clusters table in the clusters content outputs"""
        settings = self.settings
        with self.clusters_content_out:
            clear_output(wait=True)

            # This call will trigger calculating the cluster list if it doesn't exist
            n_clusters = len(self.settings.cluster_list)
            print(f"Number of clusters: {n_clusters:d}")
            print(settings.clusters_table())

    @utils.disable_cls_widget("view_clusters_button")
    @update_statusbar
    def _on_view_clusters_click(self, b) -> None:
        """Event handler when view clusters button is clicked."""
        logger.info("Temporarily disabled.")
        return
        with self.event_context(logger=logger):
            logger.info("Opening clusters in the ASE GUI")
            self.view_clusters()

    def view_clusters(self) -> None:
        """Visualize the clusters in the ASE GUI"""
        self.settings.view_clusters()



from typing import Optional
from .app_data import AppDataKeys, AppData
from .main_dashboard import MainDashboard
from .singleton import Singleton

__all__ = ["CLEASEGui", "BaseGUI", "display_ui"]


class BaseGUI:
    """The base version of the CLEASE GUI. Not a singleton, so it is better for testing."""

    def __init__(self, app_data: Optional[dict] = None, dev_mode: bool = False):

        if app_data is None:
            # The default app data.
            app_data = AppData(**{AppDataKeys.DEV_MODE: dev_mode})
        # We "write protect" these with property tags.
        self._app_data = app_data
        self._dashboard = MainDashboard(self.app_data)

    @property
    def app_data(self) -> dict:
        return self._app_data

    @property
    def dashboard(self) -> MainDashboard:
        return self._dashboard

    def display(self) -> None:
        """Display the GUI in a notebook"""
        self.dashboard.display()

    @property
    def dev_mode(self) -> bool:
        return self.app_data.get(AppDataKeys.DEV_MODE, False)

    def __repr__(self) -> str:
        # This gets printed if it's the last line in the Jupyter cell,
        # so we don't want to populate it with too much for a regular user.
        cls_name = self.__class__.__name__
        extra_info = ""
        if self.dev_mode:
            extra_info += f"{self.app_data}"
        return f"{cls_name}({extra_info})"


class CLEASEGui(BaseGUI, metaclass=Singleton):
    """The main CLEASE GUI class.
    It is a singleton, so repeated calls to display_ui returns the exact same instance."""

    # This class is a separate class from the CLEASEGui class, in order to allow
    # non-singleton version of the test app.


def display_ui(**kwargs) -> CLEASEGui:
    """The main CLEASE GUI function.
    Add the following lines in a Jupyter notebook cell:

    from clease_gui import display_ui
    display_ui()
    """

    app = CLEASEGui(**kwargs)
    app.display()
    return app



from clease_gui.version import __version__, version_info
from clease_gui.logging_widget import register_logger
from clease_gui.app_data import AppDataKeys
from clease_gui.main import display_ui, CLEASEGui

# Things which are promoted to be importable directly with
# from clease_gui import ...
__all__ = [
    "CLEASEGui",
    "display_ui",
    "register_logger",
    "version_info",
    "__version__",
    "AppDataKeys",
]



from .canonical_mc import *
from .plot_mc_page import *
from .mc_run_dashboard import *
from .view_mc_dashboard import *
from .mc_main import *

__all__ = (
    canonical_mc.__all__
    + mc_main.__all__
    + plot_mc_page.__all__
    + mc_run_dashboard.__all__
    + view_mc_dashboard.__all__
)



import logging
from IPython.display import display
import ipywidgets as widgets

from clease.settings import ClusterExpansionSettings
from clease_gui import utils
from clease_gui.base_dashboard import BaseDashboard
import clease_gui as gui
from . import settings_buttons as swidgets

__all__ = ["StructureSettingsDashboard"]

logger = logging.getLogger(__name__)
gui.register_logger(logger)

DESC_WIDTH = "150px"
STYLE = {"description_width": DESC_WIDTH}


class StructureSettingsDashboard(BaseDashboard):
    """Dashboard related to the structure settings, i.e.
    CEBulk, CECrystal and CESlab
    """

    def initialize(self):
        # Buttons related to the settings type,
        # CEBulk, CECrystal and CESlab

        # Output object for type buttons
        self.out_type = widgets.Output()
        self.out_help = widgets.Output()

        self.type_buttons_dct = {
            "CEBulk": swidgets.CEBulkButtons(self.app_data, output=self.out_type),
            "CECrystal": swidgets.CECrystalButtons(self.app_data, output=self.out_type),
            "CESlab": swidgets.CESlabButtons(self.app_data, output=self.out_type),
        }
        self.type_mode_b = widgets.Dropdown(
            options=[
                ("Bulk", "CEBulk"),
                # Not yet supported
                # ('Crystal', 'CECrystal'),
                # ('Slab', 'CESlab'),
            ],
            description="Type:",
            layout={"width": "max-content"},
            style=STYLE,
        )

        # Add observer to the type, so we know if we need to update the button collection
        def on_type_mode_change(change):
            if utils.is_value_change(change):
                self.update_type_mode()

        self.type_mode_b.observe(on_type_mode_change)

    def update_type_mode(self):
        self.type_buttons.display_widgets()

    @property
    def type_buttons(self):
        mode = self.type_mode_b.value
        return self.type_buttons_dct[mode]

    def display(self):
        super().display()
        display(self.type_mode_b)
        display(self.out_type)
        self.update_type_mode()

    @property
    def type_mode(self):
        return self.type_mode_b.value

    def get_type_kwargs(self):
        return self.type_buttons.value

    def get_settings_kwargs(self):
        kwargs = dict(
            type=self.type_mode,
            **self.get_type_kwargs(),
        )
        return kwargs

    def set_widgets_from_load(self, settings: ClusterExpansionSettings) -> None:
        # Figure out what the builder was
        kwargs = settings.kwargs
        factory = kwargs.get("factory", None)
        if factory != "CEBulk":
            logger.debug("Cannot load settings of factory type: %s", factory)
            return

        # Let's try and load things
        self._load_cebulk_settings(settings)

    def _load_cebulk_settings(self, settings: ClusterExpansionSettings) -> None:
        mode = "CEBulk"
        self.type_mode_b.value = mode
        self.update_type_mode()

        # Get some factory keys
        for key in {"crystalstructure", "a", "c", "c_over_a"}:
            value = settings.kwargs.get(key, None)
            if value is not None:
                self.type_buttons.set_widget_value(key, value)



import ipywidgets as widgets

__all__ = ["WidgetStyleMixin"]

STYLE = {"description_width": "150px"}
LAYOUT = {"width": "max-content"}


class WidgetStyleMixin:
    DEFAULT_STYLE_KWARGS = {"style": STYLE, "layout": LAYOUT}
    DEFAULT_BUTTON_LAYOUT = widgets.Layout(**LAYOUT)



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
import pandas as pd
import ase
import numpy as np

from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar
from clease_gui import register_logger
import clease_gui.utils as utils

__all__ = ["DBViewerDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class DBViewerDashboard(BaseDashboard):
    def initialize(self):

        self.db_select_widget = widgets.Text(description="DB name:", value="")
        db_name_help = widgets.Label(
            ("Leave empty to use database specified in the settings object.")
        )

        self.db_select_box = widgets.HBox(children=[self.db_select_widget, db_name_help])

        self.selection_parameters_widget = widgets.Text(description="Selection:", value="")

        selection_help_label = widgets.Label(
            (
                "For more information on querying, see "
                "https://wiki.fysik.dtu.dk/ase/ase/db/db.html#querying"
            )
        )
        self.selection_box = widgets.HBox(
            children=[self.selection_parameters_widget, selection_help_label]
        )

        self.draw_selection_button = widgets.Button(description="Open in ASE GUI")
        self.draw_selection_button.on_click(self._on_draw_selection_click)

        self.print_db_button = widgets.Button(description="Show DB content")
        self.print_db_button.on_click(self._on_print_db)

        # Output for displaying the ASE database as a pandas DF
        self.db_content_out = widgets.Output(
            layout=dict(
                overflow="auto",
                width="100%",
                height="350px",
            )
        )
        self.max_rows_widget = widgets.BoundedIntText(
            value=50,
            min=0,
            max=99999,
            description="Rows:",
        )
        self.max_rows_help = widgets.Label(
            "Maximum number of rows to display in the table (0 for unlimited)."
        )
        self.count_rows_btn = utils.make_clickable_button(
            self._on_count_rows_click, description="Count Selection"
        )
        self.count_label = widgets.Label(value=None)

    @update_statusbar
    @utils.disable_cls_widget("count_rows_btn")
    def _on_count_rows_click(self, b):
        with self.event_context(logger=logger):
            n_rows = self._count_rows()
            self.count_label.value = f"Number of rows: {n_rows}"
            logger.info("Number of rows with selection: %d", n_rows)

    def _count_rows(self):
        con = self.get_connection()
        selection = self.get_selection_parameters()
        return con.count(selection=selection)

    @update_statusbar
    @utils.disable_cls_widget("print_db_button")
    def _on_print_db(self, b):
        with self.event_context(logger=logger):
            self._print_db()

    def _print_db(self):
        db_name = str(self.db_name)
        selection = self.get_selection_parameters()
        con = self.get_connection()

        tot = self._count_rows()
        out = self.db_content_out
        logger.info("Querying database %s for %d entries", db_name, tot)

        rows = con.select(selection)
        df = ase_db_rows_to_df(rows, nmax=self.max_rows_widget.value)
        with out:
            clear_output(wait=True)
            with pd.option_context("display.max_rows", None, "display.max_columns", None):
                # Temporarily set display context to all of the rows
                # and columns we have.
                # We don't wanna truncate here.
                display(df)

    def display(self):
        buttons = widgets.HBox(
            children=[
                self.count_rows_btn,
                self.print_db_button,
                self.draw_selection_button,
            ]
        )
        rows_box = widgets.HBox(children=[self.max_rows_widget, self.max_rows_help])
        display(
            self.db_select_box,
            self.selection_box,
            rows_box,
            buttons,
            self.count_label,
            self.db_content_out,
        )

    @property
    def db_name(self):
        value = self.db_select_widget.value
        if value == "":
            # Use DB name from settings
            return self.get_db_name()
        return self.app_data[self.KEYS.CWD] / value

    @property
    def settings(self):
        return self.app_data[self.KEYS.SETTINGS]

    def get_connection(self):
        return ase.db.connect(self.db_name)

    def get_selection_parameters(self):
        value = self.selection_parameters_widget.value
        if value == "":
            return None
        return value

    def get_selection(self):
        selection = self.get_selection_parameters()
        con = self.get_connection()

        return [row.toatoms() for row in con.select(selection=selection)]

    @update_statusbar
    @utils.disable_cls_widget("draw_selection_button")
    def _on_draw_selection_click(self, b) -> None:
        with self.event_context(logger=logger):
            self._draw_selection()

    def _draw_selection(self) -> None:
        from ase.visualize import view

        images = self.get_selection()
        # Remove any calculator, to remove energies, and avoid an
        # annoying error.
        # We need to keep using the ase.visualize.view, in order to have
        # it run as a subprocess (multiprocessing & threading don't work :( )
        for image in images:
            image.calc = None
        view(images)


def ase_db_rows_to_df(rows, nmax=0) -> pd.DataFrame:
    """Convert an ASE database into a pandas data frame"""
    # rows = list(rows)
    base_keys = [
        "id",
        "formula",
        "calculator",
        "energy",
        "natoms",
        "pbc",
        "volume",
        "charge",
        "mass",
    ]

    logger.debug("Reading data from database.")
    rows_list = []
    for ii, row in enumerate(rows):
        if nmax > 0 and ii > nmax:
            break
        data = {}
        for key in base_keys:
            data[key] = getattr(row, key, None)
        kvp = row.key_value_pairs
        data.update(**kvp)
        rows_list.append(data)

    logger.debug("Parsing keys...")

    # Get all keys, ensure base keys are sorted in their order
    all_keys = list(base_keys)
    all_keys_set = {key for data in rows_list for key in data}
    for key in all_keys_set:
        if key not in all_keys:
            all_keys.append(key)
    logger.debug("Constructing pandas dataframe.")
    return pd.DataFrame(rows_list, columns=all_keys).replace({np.nan: None})


def _formatter(key, value):
    """Helper function to format the value, possibly
    on the basis of the name of the key"""
    if value is None:
        value = "-"
    if key == "pbc":
        # Convert [True, True, True] => 'TTT'
        # or [True, False, False] => 'TFF'
        value = "".join([str(v)[0] for v in value])

    return value



"""This dashboard is for the plots"""
import logging
import matplotlib.pyplot as plt
from IPython.display import display, clear_output
import ipywidgets as widgets

import clease.plot_post_process
from clease_gui import register_logger
from clease_gui.status_bar import update_statusbar
from clease_gui.base_dashboard import BaseDashboard
import clease_gui as gui

__all__ = ["FitPlotDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class FitPlotDashboard(BaseDashboard):
    def initialize(self):

        tabs = []
        titles = []

        plt.ioff()

        def add_tab(output, title):
            tabs.append(output)
            titles.append(title)
            with output:
                clear_output()

        self.fit_out = widgets.Output()
        add_tab(self.fit_out, "Fit")

        self.residual_out = widgets.Output()
        add_tab(self.residual_out, "Residuals")

        self.eci_out = widgets.Output()
        add_tab(self.eci_out, "ECI")

        self.cv_out = widgets.Output()
        add_tab(self.cv_out, "CV")

        self.convex_hull_out = widgets.Output()
        add_tab(self.convex_hull_out, "Convex Hull")

        self.plot_btn = widgets.Button(description="Update Plots", button_style="primary")

        self.plot_btn.on_click(self._on_make_plot_click)

        self.tabs = widgets.Tab(children=tabs)
        for ii, title in enumerate(titles):
            self.tabs.set_title(ii, title)

        # Keep track of open figures, so we can close them again
        # if we overwrite them.
        self._figures = {}

    def display(self):
        display(self.plot_btn, self.tabs)

    @property
    def evaluator(self):
        try:
            return self.app_data[self.KEYS.EVALUATE]
        except KeyError:
            raise KeyError("No fit has been performed yet.") from None

    def _close_figure(self, name: str) -> None:
        """Close a figure object with a certain name
        if it exists"""
        if name in self._figures:
            logger.debug("Closing figure: %s", name)
            fig = self._figures[name]
            plt.close(fig=fig)
            # We no longer need to store this
            self._figures.pop(name)

    def _register_figure(self, name: str, fig) -> None:
        """Register a figure object with a name.
        Will close the old figure if it exists"""
        self._close_figure(name)
        self._figures[name] = fig

    def _make_convex_hull(self):
        evaluator = self.evaluator

        fig = clease.plot_post_process.plot_convex_hull(evaluator)
        btn = self._make_save_figure_button(fig, "convex_hull.png")

        with self.convex_hull_out:
            clear_output(wait=True)
            display(fig, btn)
        self._register_figure("convex_hull", fig)

    def _make_residual_plot(self):
        evaluator = self.evaluator
        fig = clease.plot_post_process.plot_fit_residual(evaluator)
        btn = self._make_save_figure_button(fig, "fit_residual.png")

        with self.residual_out:
            clear_output(wait=True)
            display(fig, btn)
        self._register_figure("residual", fig)

    def _make_cv_plot(self):
        evaluate = self.evaluator
        try:
            fig = clease.plot_post_process.plot_cv(evaluate)
            btn = self._make_save_figure_button(fig, "fit_cv.png")
        except ValueError:
            # No CV scores
            fig = None
            btn = None

        with self.cv_out:
            clear_output(wait=True)
            if fig is not None:
                display(fig, btn)
                self._register_figure("cv_score", fig)
            else:
                print("No CV values, requires alpha optimization")

    def _make_fit_plot(self):
        evaluate = self.evaluator
        fig = clease.plot_post_process.plot_fit(evaluate)
        btn = self._make_save_figure_button(fig, "fit_plot.png")

        with self.fit_out:
            clear_output(wait=True)
            display(fig, btn)
        self._register_figure("fit_plot", fig)

    def _make_eci_plot(self):
        evaluate = self.evaluator
        fig = clease.plot_post_process.plot_eci(evaluate)
        btn = self._make_save_figure_button(fig, "fit_eci.png")

        with self.eci_out:
            clear_output(wait=True)
            display(fig, btn)
        self._register_figure("eci_plot", fig)

    def _make_save_figure_button(self, fig, default_filename):
        button = widgets.Button(description="Save Figure")
        filename_widget = widgets.Text(description="Filename:", value=default_filename)

        def _on_click(b):
            cwd = self.app_data[self.KEYS.CWD]
            filename = filename_widget.value
            pa = cwd / filename
            fig.savefig(
                filename,
                dpi=400,
                bbox_inches="tight",
                facecolor="white",
                transparent=False,
            )
            logger.info("Saved figure to file: %s", str(pa))

        button.on_click(_on_click)
        box = widgets.HBox(children=[button, filename_widget])
        return box

    @update_statusbar
    @gui.utils.disable_cls_widget("plot_btn")
    def _on_make_plot_click(self, b):
        with self.event_context(logger=logger):
            self._make_fit_plot()
            self._make_residual_plot()
            self._make_eci_plot()
            self._make_cv_plot()
            self._make_convex_hull()
            logger.info("Plotting completed.")

    @property
    def fit_results(self):
        return self.app_data.get("fit_results", None)



from pathlib import Path
import logging
from collections import namedtuple
from IPython.display import display, clear_output
import ipywidgets as widgets

import clease_gui.utils as utils
from clease_gui.logging_widget import (
    default_handler,
    set_all_levels,
    register_logger,
    initialize_clease_gui_logging,
)
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.app_data import save_app_data, load_app_data
from clease_gui.status_bar import update_statusbar

__all__ = ["MainDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class TestAppData(BaseDashboard):
    def initialize(self):
        self.test_button = widgets.Button(description="See app data")
        self.test_button.on_click(self._on_click)
        self.output = widgets.Output()

    def display(self):
        display(self.test_button, self.output)

    def _on_click(self, b):
        from pprint import pprint

        data = self.app_data.copy()
        if "settings" in data:
            data["settings"] = data["settings"].todict()
        with self.output:
            clear_output(wait=True)
            pprint(data)


class MainDashboard(BaseDashboard):
    def initialize(self):
        # We need to delay the import of these packages, to break cyclic imports
        from .status_bar import StatusBar
        from .db_viewer import DBViewerDashboard
        from .settings_maker import SettingsMakerDashboard
        from .new_structures import NewStructureDashboard
        from .calculate import CalculateDashboard
        from .fitting import FittingDashboard
        from .supercell import SupercellDashboard
        from .montecarlo import MainMCDashboard

        # The status bar should be the first thing to be initialized,
        # since it adds itself to the app_data
        self.status_bar = StatusBar(self.app_data)
        self.status_output = widgets.Output()
        with self.status_output:
            clear_output()
            self.status_bar.display()

        Tab = namedtuple("Tab", "output,name")
        tabs = []  # List containing the tab objects to be added
        self.dashboards = {}  # Store the dashboards for internal access

        def add_tab(dashboard, display_name):
            """Helper function to add new tabs to the main dashboard"""
            output = widgets.Output()
            # Create the dashboard
            board = dashboard(self.app_data)
            name = display_name.lower().replace(" ", "_")
            self.dashboards[name] = board
            with output:
                clear_output()
                board.display()

            tab = Tab(output, display_name)
            tabs.append(tab)

        # Add main tabs
        add_tab(SettingsMakerDashboard, "Settings")
        add_tab(NewStructureDashboard, "New Structures")
        add_tab(DBViewerDashboard, "Inspect DB")
        add_tab(CalculateDashboard, "Calculate")
        add_tab(FittingDashboard, "ECI")
        add_tab(SupercellDashboard, "Supercell")
        add_tab(MainMCDashboard, "Monte Carlo")
        if self.dev_mode:
            add_tab(TestAppData, "Testing")

        # Add the tabs to the main dashboard
        self.tab = widgets.Tab(children=[tab.output for tab in tabs])

        # Add current working directory widet in the top
        cwd = Path(".").resolve()
        self.cwd_widget = widgets.Text(
            value=str(cwd),
            description="Current working directory:",
            layout=widgets.Layout(height="auto", width="100%"),
            style={"description_width": "initial"},
        )
        # Update the app data, and observe any changes
        self.set_cwd(self.cwd_widget.value)
        self.cwd_widget.observe(self._on_cwd_change)
        self.logo = _get_logo_widget()

        for ii, tab in enumerate(tabs):
            self.tab.set_title(ii, tab.name)

        # Add button to clear logs
        clear = widgets.Button(description="Clear Logs")
        log_level = widgets.Dropdown(
            options=[
                ("Error", logging.ERROR),
                ("Warning", logging.WARNING),
                ("Info", logging.INFO),
                ("Debug", logging.DEBUG),
            ],
            value=logging.INFO,
            description="Logging level:",
            style={"description_width": "initial"},
        )

        def on_clear_click(b):
            default_handler.clear_logs()

        clear.on_click(on_clear_click)

        def on_log_level_change(change):
            if utils.is_value_change(change):
                set_all_levels(change["new"])

        log_level.observe(on_log_level_change)
        initialize_clease_gui_logging()

        self.log_box = widgets.HBox(children=[clear, log_level])

        # save/load app state button
        self.save_app_state_button = widgets.Button(description="Save app data")
        self.save_app_state_button.on_click(self._on_save_app_data)
        self.load_app_state_button = widgets.Button(description="Load app data")
        self.load_app_state_button.on_click(self._on_load_app_data)

    def display(self):
        # Top bar box with logo and CWD widget
        top_bar_box = widgets.HBox(
            children=[self.logo, self.cwd_widget], layout={"align_items": "center"}
        )
        display(top_bar_box)
        display(self.tab)
        display(self.status_output)
        default_handler.show_logs()
        display(self.log_box)
        app_data_state_buttons = widgets.HBox(
            children=[self.save_app_state_button, self.load_app_state_button]
        )
        display(app_data_state_buttons)

    @property
    def cwd(self):
        return self.app_data[self.KEYS.CWD]

    def set_cwd(self, value):
        self.app_data[self.KEYS.CWD] = Path(value).resolve()

    def _on_cwd_change(self, change) -> Path:
        if utils.is_value_change(change):
            self.set_cwd(change.new)

    @property
    def app_data_file_name(self):
        # Hardcode app data filename for now.
        return self.cwd / ".clease_gui_app_data.json"

    @update_statusbar
    def _on_save_app_data(self, b):
        with self.event_context(logger=logger):
            fname = self.app_data_file_name
            logger.info("Saving app data to file: %s", str(fname))
            self._save_app_data(fname)
            logger.info("Save complete.")

    def _save_app_data(self, filename):
        save_app_data(self.app_data, filename)

    def _load_app_data(self, filename):
        new_data = load_app_data(filename)
        # Get the supercell, and make the supercell dashboard set it
        supercell = self.app_data.get(self.KEYS.SUPERCELL, None)
        if supercell is not None:
            logger.debug("Updating supercell from app data")
            dashboard = self.dashboards["supercell"]
            dashboard.set_supercell(supercell)

        self.app_data.update(new_data)
        # Try to update the settings page
        settings = new_data.get(self.KEYS.SETTINGS, None)
        if settings is not None:
            settings_dashboard = self.dashboards["settings"]
            settings_dashboard.set_settings_widget_states(settings)

    @update_statusbar
    def _on_load_app_data(self, b):
        with self.event_context(logger=logger):
            fname = self.app_data_file_name
            logger.info("Loading app data from file: %s", str(fname))
            self._load_app_data(fname)
            logger.info("Load complete.")


def _get_logo_widget():
    """Function to create the CLEASE logo and return it in an image widget"""
    logo_path = utils.get_assets_path() / "clease_logo.png"
    with logo_path.open("rb") as file:
        logo = file.read()
    return widgets.Image(
        value=logo,
        format="png",
        width=80,
    )



import logging
import uuid
from pathlib import Path
from IPython.display import display
import ipywidgets as widgets
import json
from clease_gui import register_logger, utils
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar

__all__ = ["MCRunDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class MCRunDashboard(BaseDashboard):
    def initialize(self):
        s = "Export MC run data to JSON file for further analysis, "
        s += "or load MC run data from file."
        html = f"<h1 style='font-size:15px'> {s} </h1>"
        self.description = widgets.HTML(html)
        self.filename_widget = widgets.Text(
            description="Filename:", value="clease_mc.json", **self.DEFAULT_STYLE_KWARGS
        )

        self.mode_widget = widgets.Dropdown(
            description="MC mode:",
            options=[
                ("Canonical", "canonical"),
            ],
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.save_button = widgets.Button(description="Export data")
        self.save_button.on_click(self._on_save_click)

        self.append_widget = widgets.Checkbox(
            description="Append data in save/load?",
            value=True,
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.load_data_button = widgets.Button(description="Load Data")
        self.load_data_button.on_click(self._on_load_click)

    def display(self):
        hbox = widgets.HBox(children=[self.save_button, self.load_data_button])
        display(
            self.description,
            self.mode_widget,
            self.filename_widget,
            self.append_widget,
            hbox,
        )

    def get_filename(self):
        cwd = self.get_cwd()
        return cwd / self.filename_widget.value

    @property
    def mc_mode(self):
        return self.mode_widget.value

    def get_data(self):
        """Retrieve the MC run data from the app data"""
        mode = self.mc_mode
        if mode == "canonical":
            key = self.KEYS.CANONICAL_MC_DATA
        else:
            raise ValueError(f"Unknown mode: {mode}")

        try:
            return list(self.app_data[key])
        except KeyError:
            raise KeyError("No data found. Did you run the MC yet?")

    def set_data(self, data, append=False) -> None:
        """Set the data of the currently active MC mode. If append=False, it will
        override any existing runs, and otherwise it will only append runs which
        aren't already in the list, i.e. the UUID does not exist in the list."""
        # We shouldn't have UUID collisions, but just in case we made a mistake.
        assert len({get_uuid_from_run(run) for run in data}) == len(
            data
        ), "Not all runs have unique ID's"
        mode = self.mc_mode
        if mode == "canonical":
            key = self.KEYS.CANONICAL_MC_DATA
        else:
            raise ValueError(f"Unknown mode: {mode}")

        if append:
            all_data = self.app_data.get(key, [])
            # Iterate through incoming new data, only append the runs which are new
            uids = {get_uuid_from_run(run) for run in all_data}
            for run in data:
                new_id = get_uuid_from_run(run)
                if new_id in uids:
                    # We already have this run
                    continue
                all_data.append(run)
        else:
            # Just override all the data
            all_data = data
        self.app_data[key] = all_data
        logger.debug("Updated MC data for mode: %s", mode)

    def save_data(self, filename):
        data = self.get_data()
        data = self.append_data(data, filename)
        with open(filename, "w") as file:
            json.dump(data, file, indent=4)

    @update_statusbar
    @utils.disable_cls_widget("save_button")
    def _on_save_click(self, b):
        with self.event_context(logger=logger):
            fname = self.get_filename()
            logger.info("Exporting data to file: %s", fname)
            self.save_data(fname)

    def append_data(self, data, filename):
        if not self.append_widget.value:
            return data
        if not Path(filename).is_file():
            return data

        logger.info("Found existing data file. Appending current data.")
        # Read the data, and append ours
        with open(filename) as file:
            existing_data = json.load(file)

        if not isinstance(existing_data, list):
            raise TypeError(
                (
                    f'Loaded data from file "{filename}", '
                    "which was not of the expected type."
                    f"Got {type(existing_data)}, expected list"
                )
            )

        runs_to_save = set(get_uuid_from_run(run) for run in data)
        # figure out what data we already saved
        for run in existing_data:
            uid = get_uuid_from_run(run)
            # Remove the ID from the set if it exists
            # since we already have it
            runs_to_save.discard(uid)

        # Figure out runs to save
        data = [run for run in data if get_uuid_from_run(run) in runs_to_save]

        data = existing_data + data
        return data

    def load_data(self, filename):
        with open(filename) as file:
            data = json.load(file)

        self.set_data(data, append=self.append_widget.value)

    @update_statusbar
    @utils.disable_cls_widget("load_data_button")
    def _on_load_click(self, b):
        with self.event_context(logger=logger):
            fname = self.get_filename()
            logger.info("Loading data from file: %s", fname)
            self.load_data(fname)
            logger.info("Data loaded succesfully.")


def get_uuid_from_run(run):
    meta = run.get("meta", {})
    uid = meta.get("run_id", None)
    if uid is not None:
        uid = uuid.UUID(uid)
    return uid



import logging

import threading
from IPython.display import display, clear_output
import ipywidgets as widgets
from clease.montecarlo.observers import MoveObserver

from clease_gui import register_logger, utils
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar

import nglview as nv

from .mc_nglviewer import MCNGLDisplay

__all__ = ["MCViewDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class MCViewDashboard(BaseDashboard):
    def initialize(self) -> None:
        self.viewer = MCNGLDisplay()

        self.center_btn = utils.make_clickable_button(self._center_click, description="Center")
        self.update_btn = utils.make_clickable_button(self._update_click, description="Update")
        self.play_mc = utils.make_clickable_button(self._play_mc_click, description="Play")
        self.stop_movie_btn = utils.make_clickable_button(self._stop_mc_click, description="Stop")
        self.steps_per_frame = widgets.BoundedIntText(
            min=1,
            max=999999,
            step=1,
            value=10,
            description="Steps per frame",
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.time_per_frame = widgets.BoundedFloatText(
            min=0.1,
            max=99,
            step=0.1,
            description="Time per frame (s)",
            value=0.2,
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.movie_progress = widgets.IntProgress()
        self.movie_progress_label = widgets.Label()

        self.viewer_out = widgets.Output()
        self._event = threading.Event()

        with self.viewer_out:
            clear_output()
            display(self.viewer.viewer.display(gui=False))

    def get_move_obs(self) -> MoveObserver:
        obs = self.app_data.get(self.KEYS.STEP_OBSERVER, None)
        if obs is None:
            raise ValueError(
                "No MC movie has been made yet. "
                "Enable step recording in the 'Observe Steps' box "
                "on the MC page."
            )
        return obs

    def _stop_mc_click(self, b):
        self._event.set()

    def _center_click(self, b):
        self.get_ngl_widget().center()

    def display(self) -> None:
        experimental_label = widgets.Label("Note: This is still experimental")
        experimental_label.value += "\n"
        display(experimental_label)
        display(self.center_btn)
        box = widgets.HBox(children=[self.update_btn, self.steps_per_frame, self.time_per_frame])
        display(box)
        box = widgets.HBox(children=[self.play_mc, self.stop_movie_btn])
        display(box)
        box = widgets.HBox(children=[self.movie_progress, self.movie_progress_label])
        display(box)
        display(self.viewer_out)

    def handle_resize(self) -> None:
        self.get_ngl_widget().handle_resize()

    def get_ngl_widget(self) -> nv.NGLWidget:
        """Get the NGL widget object"""
        return self.viewer.viewer

    @update_statusbar
    def _play_mc_click(self, b):
        with self.event_context(logger=logger):
            self._play_mc_in_thread()

    @utils.disable_cls_widget("update_btn")
    def _update_click(self, b):
        with self.event_context(logger=logger):
            self._update()

    def _update(self):
        """Construct a new NGL widget, and re-draw it"""
        move_obs = self.get_move_obs()
        self.viewer.make_viewer(move_obs.base_atoms)
        self.draw_viewer()

    def draw_viewer(self):
        """Redraw the MC Viewer widget"""
        with self.viewer_out:
            clear_output(wait=True)
            self.viewer.display()

    def _play_mc_in_thread(self):
        self.get_move_obs()  # Ensure the move observer exists

        # Remake the viewer
        thread_draw = threading.Thread(target=self._update, daemon=True)
        thread_draw.start()
        thread_draw.join()
        # Start the movie
        thread = threading.Thread(target=self._play_mc, daemon=True)
        thread.start()

    @utils.disable_cls_widget("play_mc")
    @utils.disable_cls_widget("update_btn")
    def _play_mc(self):
        """Start the movie"""
        move_obs = self.get_move_obs()
        self._event.clear()

        logger.info("Playing movie")
        tot = len(move_obs.steps)
        self.movie_progress.max = tot
        self.movie_progress.value = 0

        def _do_update(ii, atoms):
            self.movie_progress.value = ii
            self.movie_progress_label.value = f"{ii}/{tot}"
            self.viewer.replace_structure(atoms)

        # for ii, step in enumerate(move_obs.steps, start=1):
        for ii, atoms in enumerate(move_obs.reconstruct_iter(), start=1):
            if self._event.is_set():
                logger.info("Stopping movie.")
                self.movie_progress_label.value += ", interrupted."
                break
            if ii % self.steps_per_frame.value == 0:
                _do_update(ii, atoms)
                # Slight delay per frame
                self._event.wait(self.time_per_frame.value)
        else:
            # Movie done, no interrupt.
            # Do a final update
            _do_update(ii, atoms)
            self.movie_progress_label.value += ", done!"
            logger.info("Movie complete.")



import logging
from time import perf_counter  # Import directly for faster access in main MC loop
from datetime import timedelta
import uuid
from IPython.display import display, clear_output
import ipywidgets as widgets
import numpy as np
import ase
import clease
from clease.montecarlo import constraints
from clease_gui import register_logger, utils
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar
from clease_gui.utils import make_clickable_button
from clease_gui.timing import est_runtime_delta
from clease.montecarlo import observers

__all__ = ["CanonicalMC"]

logger = logging.getLogger(__name__)
register_logger(logger)

# Register the clease MC logger specifically.
# Allows to follow the internal messages from the MC object.
# clease_mc_logger = logging.getLogger('clease.montecarlo.montecarlo')
# register_logger(clease_mc_logger)


class CanonicalMC(BaseDashboard):
    def __init__(self, *args, **kwargs):
        self._mc = None  # MC process
        super().__init__(*args, **kwargs)

        self.step_move_obs = None

    def initialize(self):
        # Temperature widgets

        self.temperature_mode_widget = widgets.Dropdown(
            description="MC mode:",
            options=[
                ("Fixed Temperature", "temp_fixed"),
                ("Simulated annealing (linear spacing)", "temp_range_linspace"),
                ("Simulated annealing (logarithmic spacing)", "temp_range_logspace"),
            ],
            value="temp_range_logspace",
            **self.DEFAULT_STYLE_KWARGS,
        )

        temp_range_linear = make_spacing_boxes(
            style_kwargs=self.DEFAULT_STYLE_KWARGS,
        )

        temp_range_log = make_spacing_boxes(
            style_kwargs=self.DEFAULT_STYLE_KWARGS,
        )

        self.temp_widgets = {
            "temp_fixed": widgets.BoundedFloatText(
                value=300,
                min=0,
                max=9999999,
                description="Temperature:",
                **self.DEFAULT_STYLE_KWARGS,
            ),
            "temp_range_linspace": temp_range_linear,
            "temp_range_logspace": temp_range_log,
        }
        self.temp_mode_output = widgets.Output()
        self._update_temp_mode()

        def _on_temp_mode_change(change):
            """Update the steps_output display whenever we change the dropdown"""
            if utils.is_value_change(change):
                self._update_temp_mode()

        self.temperature_mode_widget.observe(_on_temp_mode_change)

        # MC logging frequency
        self.update_freq_widget = widgets.BoundedIntText(
            value=500,
            min=100,
            max=9999999,
            description="Update freq. (ms):",
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.update_freq_help_msg = widgets.Label(value="Min frequency of MC status updates.")

        self.step_mode_help = widgets.Label(
            "1 sweep runs equals <number of atoms in supercell> steps."
        )

        # Number of steps
        self.step_widgets = {}
        sweeps_widget = widgets.BoundedIntText(
            description="Number of sweeps:",
            value=1,
            min=1,
            max=99999999,
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.step_widgets["sweeps"] = sweeps_widget

        steps_widget = widgets.BoundedIntText(
            description="Number of steps:",
            value=100,
            min=1,
            max=999999999999999,
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.step_widgets["steps"] = steps_widget

        self.steps_output = widgets.Output()
        # Dropdown for selecting the step mode
        self.step_mode_widget = widgets.Dropdown(
            options=[
                ("Sweeps", "sweeps"),
                ("Steps", "steps"),
            ],
            description="Step mode:",
            value="sweeps",
            **self.DEFAULT_STYLE_KWARGS,
        )
        self._update_step_mode_output()  # update the steps_output

        def _on_step_mode_change(change):
            """Update the steps_output display whenever we change the dropdown"""
            if utils.is_value_change(change):
                self._update_step_mode_output()

        self.step_mode_widget.observe(_on_step_mode_change)

        self.run_mc_button = make_clickable_button(
            self._on_run_mc_click, description="Run MC", button_style="primary"
        )

        self.update_atoms_after_run_widget = widgets.Checkbox(
            description="Update supercell after MC run?", value=True
        )

        self.clear_mc_app_data_button = widgets.Button(description="Clear MC data")
        self.clear_mc_app_data_button.on_click(self._on_clear_mc_data)

        # Button for stopping the MC process
        self.stop_mc_button = widgets.Button(description="Abort MC", button_style="warning")
        self.stop_mc_button.on_click(self._on_stop_mc_click)

        self.mc_progress = widgets.IntProgress()
        self.mc_progress_label = widgets.Label()

        self.constrain_sublattices_wdgt = widgets.Checkbox(
            value=True,
            description="Constrain swaps by sublattice?",
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.step_observer_wdgt = widgets.Dropdown(
            options=[
                ("Off", "off"),
                # Values for the "only_accept" parameter
                ("Every step", False),
                ("Accepted steps", True),
            ],
            description="Observe steps:",
            value="off",
            **self.DEFAULT_STYLE_KWARGS,
        )

    def attach_constraints(
        self,
        mc: clease.montecarlo.Montecarlo,
        settings: clease.settings.ClusterExpansionSettings,
    ) -> None:
        if self.constrain_sublattices_wdgt.value:
            atoms = mc.atoms
            cnst = constraints.ConstrainSwapByBasis(atoms, settings.index_by_basis)
            mc.generator.add_constraint(cnst)

    @property
    def update_after_run(self) -> bool:
        """Do we update the supercell in app_data after
        a completed MC run?"""
        return self.update_atoms_after_run_widget.value

    def display(self):
        frq_box = widgets.HBox(
            children=[
                self.update_freq_widget,
                self.update_freq_help_msg,
            ]
        )
        step_mode_box = widgets.HBox(
            children=[
                self.step_mode_widget,
                self.step_mode_help,
            ]
        )

        # place update atoms checkbox next to run MC button
        run_mc_box = widgets.HBox(
            children=[
                self.run_mc_button,
                self.stop_mc_button,
                self.update_atoms_after_run_widget,
            ],
            layout=dict(justify_content="space-between"),
        )

        display(step_mode_box, self.steps_output)
        display(self.temperature_mode_widget, self.temp_mode_output)
        display(self.step_observer_wdgt)
        display(frq_box, self.constrain_sublattices_wdgt)
        display(self.clear_mc_app_data_button)

        display(run_mc_box)
        mc_progress_box = widgets.VBox(
            children=[
                self.mc_progress,
                self.mc_progress_label,
            ]
        )
        display(mc_progress_box)

    @property
    def active_temp_mode(self):
        return self.temperature_mode_widget.value

    @property
    def active_temp_mode_widget(self):
        return self.temp_widgets[self.active_temp_mode]

    def _update_temp_mode(self):
        with self.temp_mode_output:
            clear_output(wait=True)
            display(self.active_temp_mode_widget)

    def get_temp(self):
        mode = self.active_temp_mode
        widget = self.active_temp_mode_widget
        if mode == "temp_fixed":
            value = widget.value
            return np.array([value])
        elif mode == "temp_range_linspace":
            t_max = widget.children[0].value
            t_min = widget.children[1].value
            t_steps = widget.children[2].value
            return np.linspace(t_max, t_min, t_steps)
        elif mode == "temp_range_logspace":
            t_max = widget.children[0].value
            t_min = widget.children[1].value
            t_steps = widget.children[2].value
            # Logspace is in base 10 by default
            return np.logspace(np.log10(t_max), np.log10(t_min), t_steps)
        raise ValueError(f"Unknown mode: {mode}")

    def _get_key(self, key):
        try:
            return self.app_data[key]
        except KeyError:
            raise KeyError(f"No key named {key} in app data. Must be loaded/created first.")

    @property
    def active_step_mode(self):
        return self.step_mode_widget.value

    @property
    def active_step_mode_widget(self):
        mode = self.active_step_mode
        return self.step_widgets[mode]

    def _update_step_mode_output(self):
        with self.steps_output:
            clear_output(wait=True)
            display(self.active_step_mode_widget)

    def get_number_of_steps(self):
        mode = self.active_step_mode
        value = self.active_step_mode_widget.value
        if mode == "sweeps":
            return value * len(self.atoms)
        assert mode == "steps"
        return value

    @property
    def atoms(self) -> ase.Atoms:
        return self._get_key(self.KEYS.SUPERCELL)

    @property
    def eci(self):
        return self._get_key(self.KEYS.ECI)

    @property
    def settings(self):
        return self._get_key(self.KEYS.SETTINGS)

    @property
    def update_freq(self):
        return self.update_freq_widget.value

    def set_supercell(self, supercell):
        self.app_data[self.KEYS.SUPERCELL] = supercell

    def _make_mc_object(self, atoms, temp=200):
        eci = self.eci
        settings = self.settings

        # attach a CLEASE calculator
        atoms = clease.calculator.attach_calculator(settings, atoms, eci=eci)
        mc = clease.montecarlo.Montecarlo(atoms, temp)

        self.attach_constraints(mc, settings)

        return mc

    def _on_clear_mc_data(self, b):
        """Remove all data from previous MC runs"""
        try:
            del self.app_data[self.KEYS.CANONICAL_MC_DATA]
            logger.info("MC data has been cleared")
        except KeyError:
            logger.info("No MC data to clear.")
            # There was no data
            pass

    def set_mc_data(self, data):
        self.app_data[self.KEYS.CANONICAL_MC_DATA] = data

    @property
    def mc_data(self) -> list:
        return self._get_key(self.KEYS.CANONICAL_MC_DATA)

    def get_mc_data(self) -> list:
        """Return existing app data if it exists, otherwise return a new
        MC data list."""
        try:
            return self.mc_data
        except KeyError:
            # No data exists, create a new data dictionary
            # and insert it
            data = []
            self.set_mc_data(data)
            return self.mc_data

    def _on_stop_mc_click(self, b):
        """Set a termination flag in the currently active MC object,
        if there is one"""
        if self._mc is None:
            return
        self._mc.quit = True

    def _check_required_keys(self) -> None:
        """Check that relevant keys  for a MC run is present.
        Raises an error if that's not the case.
        """
        missing_keys = []
        for key in (self.KEYS.SUPERCELL, self.KEYS.SETTINGS, self.KEYS.ECI):
            if key not in self.app_data:
                missing_keys.append(key)
        if len(missing_keys) > 0:
            missing_keys = ", ".join(missing_keys)
            raise KeyError(f'Missing the following in app data: "{missing_keys}".')

    @update_statusbar
    def _gui_run_mc(self) -> None:
        """Start the `_run_mc` method in a separate thread,
        while also disabling the run mc button for the duration
        of the run
        """

        # First we check if we have all necessary quantities, before
        # launching the thread
        self._check_required_keys()

        # May be run in a separate process, eventually. For now, use main thread.
        with utils.disable_widget_context(self.run_mc_button):
            with self.event_context(logger=logger):
                self._run_mc()

    def _run_mc(self):
        """Execute the MC run at the currnet settings."""
        # Some initialization
        temperatures = self.get_temp()
        n_temp = len(temperatures)

        # Initial atoms object, current working atoms
        atoms = self.atoms

        def check_abort(mc):
            """Helper function to raise if we abort the MC.
            Raises an error, to break the thread which is running
            it"""
            if mc.quit:
                val = self.mc_progress_label.value
                self.mc_progress_label.value = val + ". MC aborted!"
                raise RuntimeError("MC was aborted.")

        mc = self._make_mc_object(atoms)
        # Add emin observer
        emin_obs = observers.LowestEnergyStructure(mc.atoms)
        mc.attach(emin_obs)
        self._mc = mc

        # Add move observer
        _step_val = self.step_observer_wdgt.value
        self.step_move_obs = None
        if _step_val != "off":
            self.step_move_obs = observers.MoveObserver(atoms, only_accept=_step_val)
            mc.attach(self.step_move_obs)
        self.app_data[self.KEYS.STEP_OBSERVER] = self.step_move_obs

        # Read in the number of steps.
        steps = self.get_number_of_steps()
        # Initialize the progress bar
        self.mc_progress.value = 0
        self.mc_progress.max = n_temp
        start_time = perf_counter()
        logger.info("Starting MC with %d temperature steps, and %d swaps per step.", n_temp, steps)

        # Total number of swaps which will be performed.
        tot_mc_steps = steps * n_temp

        # Template for the progress bar label string
        msg_template = f"""
        {{prog:.1f}}%, {{ii}}/{n_temp}.
        Current temp: {{temp:.1f}} K.
        MC swaps: {{cur_mc_step:d}}/{tot_mc_steps:d}.
        Runtime: {{dt}}. Remaining: {{estimated}}.
        """
        msg_template = msg_template.strip().replace("\n", " ")

        def _update_progress_label(ii, temp, mc_step=None) -> None:
            """Helper function to update the progress label."""
            runtime = perf_counter() - start_time
            dt = timedelta(seconds=int(runtime))
            # Calculate the number of swaps that have been performed, so far.
            cur_mc_step = ii * steps
            if mc_step is not None:
                cur_mc_step += mc_step.step
            # prog = (ii + 1) / n_temp * 100  # Progress in percentage
            prog = cur_mc_step / tot_mc_steps
            # Estimated time remaining.
            estimated = est_runtime_delta(prog, runtime)
            # Fill in the remaining variables.
            msg = msg_template.format(
                prog=prog * 100,
                ii=ii + 1,
                temp=temp,
                cur_mc_step=cur_mc_step,
                dt=dt,
                estimated=estimated,
            )
            self.mc_progress_label.value = msg

        last_step_time = perf_counter()
        update_freq = self.update_freq / 1000  # from ms to s
        run_data = {}
        try:
            for ii, temp in enumerate(temperatures):
                # Update progress bar
                self.mc_progress.value += 1
                _update_progress_label(ii, temp)
                # Initialize MC things
                check_abort(mc)
                mc.T = temp
                # Ensure average energies are cleared.
                _clear_energies(mc)

                initial_energy = mc.current_energy

                # some debugging logging
                logger.debug(
                    "Starting MC run at T = %.3f K. Running a total of %d MC steps.",
                    temp,
                    steps,
                )
                # Simmulated annealing
                logger.debug(
                    "Annealing: Temperature step %d of %d. Current temperature: %.3f K.",
                    ii,
                    n_temp,
                    temp,
                )

                # Execute the MC
                for mc_step in mc.irun(steps):
                    # Update the progress string every N seconds
                    if perf_counter() - last_step_time > update_freq:
                        _update_progress_label(ii, temp, mc_step=mc_step)
                        last_step_time = perf_counter()
                # One final update of the label.
                _update_progress_label(ii, temp, mc_step=mc_step)
                check_abort(mc)  # Check if we exited because of an abort
                logger.debug("MC done after %d MC steps.", steps)
                logger.debug("Energy before MC: %.3f eV.", initial_energy)
                logger.debug(
                    "Lowest energy achieved during run: %.3f eV.",
                    emin_obs.lowest_energy,
                )
                self._update_mc_data(mc, emin_obs, run_data)

            if self.update_after_run:
                # Set the supercell to the best atoms object we found.
                self.set_supercell(emin_obs.emin_atoms)
                logger.info("Updated supercell to minimum energy configuration")

            if run_data:
                # Only append if we have any data.
                mc_app_data = self.get_mc_data()
                mc_app_data.append(run_data)

            dt = timedelta(seconds=round(perf_counter() - start_time))
            dt_per_step = dt / n_temp
            logger.info("MC run complete in %s. Avg. time per temp: %s.", dt, dt_per_step)
        finally:
            self._mc = None

    @update_statusbar
    def _on_run_mc_click(self, b):
        with self.event_context(logger=logger):
            self._gui_run_mc()

    def _make_mc_data(self, mc: clease.montecarlo.Montecarlo, emin_obs) -> dict:
        """Create a dictionary with the relevant data from the MC run.
        We pass in the relevant emin observer as well, because we always want this data."""
        lowest_energy = emin_obs.lowest_energy
        temp = mc.temperature
        accept_rate = mc.num_accepted / mc.current_step
        thermo = mc.get_thermodynamic_quantities()

        mean_energy = thermo["energy"]
        heat_capacity = thermo["heat_capacity"]
        en_var = thermo["energy_var"]

        data = dict(
            emin=lowest_energy,
            mean_energy=mean_energy,
            temperature=temp,
            accept_rate=accept_rate,
            heat_capacity=heat_capacity,
            en_var=en_var,
        )
        return data

    def _update_mc_meta(self, mc, current_data):
        """Add metadata to the current data results"""
        atoms = mc.atoms
        current_meta = current_data.get("meta", {})
        meta = dict(
            natoms=len(atoms),
            formula=atoms.get_chemical_formula(),
            temp_mode=self.active_temp_mode,
        )
        current_meta.update(meta)
        if "run_id" not in current_meta:
            current_meta["run_id"] = str(uuid.uuid4())
        current_data["meta"] = current_meta

    def _update_mc_data(self, mc: clease.montecarlo.Montecarlo, emin_obs, current_data) -> None:
        """Update the current data with the new data"""
        new_data = self._make_mc_data(mc, emin_obs)
        self._update_mc_meta(mc, current_data)

        for key, value in new_data.items():
            if key not in current_data:
                # Add an empty list if this key has not been added yet
                # (emulating a defaultdict)
                current_data[key] = []
            current_data[key].append(value)


def make_spacing_boxes(min=300, max=30000, steps=25, style_kwargs=None):
    style_kwargs = style_kwargs or {}
    temp_range_min = widgets.BoundedFloatText(
        description="Temp (min):",
        min=1,
        max=99999999,
        value=min,
        **style_kwargs,
    )
    temp_range_max = widgets.BoundedFloatText(
        description="Temp. (max):",
        min=1,
        max=99999999,
        value=max,
        **style_kwargs,
    )
    temp_range_steps = widgets.BoundedIntText(
        description="Number of temp. steps:",
        min=1,
        max=99999999,
        value=steps,
        **style_kwargs,
    )

    temp_range_box = widgets.VBox(
        children=[temp_range_max, temp_range_min, temp_range_steps],
        layout=dict(justify_content="space-between"),
    )
    return temp_range_box


def _clear_energies(mc: clease.montecarlo.Montecarlo) -> None:
    """Helper function to clear the avg energies in the MC object."""
    for avger in [mc.mean_energy, mc.energy_squared]:
        if avger is not None:
            avger.clear()



"""This module defines the core for a Kwarg type widget"""
from typing import Tuple, Sequence, List
from abc import ABC, abstractmethod
from IPython.display import display, clear_output
import ipywidgets as widgets
from clease_gui.style_mixin import WidgetStyleMixin

__all__ = ["WidgetCollection", "SingleWidget"]


class SingleWidget(WidgetStyleMixin, ABC):
    def __init__(self, name: str, widget: widgets.CoreWidget, priority: int = 0):
        self.name = name
        self.widget = widget
        self.priority = priority

    def display_widget(self):
        display(self.widget)

    @property
    def value(self):
        return self.widget.value

    @value.setter
    def value(self, value):
        self.set_value(value)

    def set_value(self, value):
        self.widget.value = value


class WidgetCollection(WidgetStyleMixin, ABC):
    """The kwarg widget contains a collection of widgets
    which gets certain key - value pairs for a function.
    All relevant widgets are constructed during initialization and registered.
    """

    def __init__(self, output=None):
        self._widgets = []

        for single_widget in self.make_widgets():
            self.register_widget(single_widget)
        self.output = output

    @abstractmethod
    def make_widgets(self) -> Sequence[SingleWidget]:
        """Make the relevant widgets, and register them."""

    def register_widget(self, widget: SingleWidget) -> None:
        name = widget.name

        if name in self.get_names():
            raise RuntimeError(f"Name {name} is already registered.")

        self.widgets.append(widget)
        self.widgets.sort(key=lambda wdgt: wdgt.priority, reverse=True)

    def get_internal_widgets(self) -> Tuple[widgets.CoreWidget]:
        """Get the raw CoreWidget objects"""
        return tuple(widget.widget for widget in self.widgets)

    @property
    def widgets(self) -> List[SingleWidget]:
        """Mapping of the internal widgets"""
        return self._widgets

    def get_names(self) -> List[str]:
        return [wdgt.name for wdgt in self.widgets]

    @property
    def value(self):
        """Return the desired value(s) from the collection of widgets."""
        return {widget.name: self.get_widget_value(widget) for widget in self.widgets}

    def get_widget_names(self):
        return tuple(widget.name for widget in self.widgets)

    def get_value_by_name(self, name):
        """Find a find widget of particular name, and return its value"""
        return self.get_widget_value(self.get_widget_by_name(name))

    def get_widget_by_name(self, name):
        index = self.get_widget_index(name)
        return self.widgets[index].widget

    def get_widget_index(self, name) -> int:
        for idx, widget in enumerate(self.widgets):
            if widget.name == name:
                return idx
        raise ValueError(
            (f"Widget of name {name} not found. " f"Available widgets: {self.get_widget_names()}")
        )

    def remove_widget(self, name):
        index = self.get_widget_index(name)
        del self.widgets[index]

    def get_widget_value(self, widget: SingleWidget):
        """Function to determine the value of a particular widget"""
        return widget.value

    def display_single_widget(self, widget: SingleWidget):
        """How to display the invidual widget"""
        widget.display_widget()

    def display_all_widgets(self):
        """How to display this collection of widgets"""
        for widget in self.widgets:
            self.display_single_widget(widget)

    def display_widgets(self, output=None, clear=True, wait=True):
        """Display the internal widgets."""

        def _display():
            if clear:
                clear_output(wait=wait)
            self.display_all_widgets()

        if output is not None:
            with output:
                _display()
        elif self.output is not None:
            with self.output:
                _display()
        else:
            _display()

    def set_widget_value(self, name, value):
        """Set the value of a widget"""
        widget = self.get_widget_by_name(name)
        widget.value = value



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import savgol_filter
import attr
from clease_gui import register_logger, utils
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar

__all__ = ["PlotMCDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


@attr.mutable
class PlotUpdater:
    tab_title: str = attr.field()
    update_fnc = attr.field()
    ax = attr.field(default=None)

    def update_plot(self):
        if self.ax is None:
            raise ValueError("No ax object has been assigned yet.")
        self.update_fnc(self.ax)


class PlotMCDashboard(BaseDashboard):
    def initialize(self):
        plt.ioff()
        self.mode_widget = widgets.Dropdown(
            description="MC mode:",
            options=[
                ("Canonical", "canonical"),
            ],
            **self.DEFAULT_STYLE_KWARGS,
        )
        plot_tabs = []
        self.plot_outputs = []
        self.figures = []

        self.plot_updaters = [
            PlotUpdater("Min. Energy", self._update_emin_plot),
            PlotUpdater("Avg. Energy", self._update_emean_plot),
            PlotUpdater("Accept Rate", self._update_accept_rate_plot),
            PlotUpdater("Heat Capacity", self._update_heat_capacity_plot),
            PlotUpdater("Heat Capacity alt.", self._update_heat_capacity_by_diff_plot),
            PlotUpdater("Energy Variance", self._update_energy_variance_plot),
        ]

        for updater in self.plot_updaters:
            tab = widgets.Output()
            out = widgets.Output()
            with out:
                clear_output()
                with plt.style.context("seaborn"):
                    fig, ax = plt.subplots()
            self.plot_outputs.append(out)
            plot_tabs.append(tab)
            self.figures.append(fig)
            updater.ax = ax  # Save the axis object in the updater

            # Create save fig button/text
            box = self._make_savefig_box(fig)
            with tab:
                clear_output()
                display(out, box)

        self.update_button = widgets.Button(description="Update")
        self.update_button.on_click(self._on_update_click)

        self.x_axis_style_widget = widgets.Dropdown(
            description="x axis scale:",
            options=[
                ("Linear", "linear"),
                ("Logarithmic", "log"),
            ],
            value="log",
        )
        self.x_axis_style_widget.observe(self._on_x_axis_change)
        self.plots_tabs = widgets.Tab(children=plot_tabs)
        for ii, updater in enumerate(self.plot_updaters):
            self.plots_tabs.set_title(ii, updater.tab_title)

        self.invert_temp_axis = widgets.Checkbox(
            value=True,
            description="Reverse temperature axis?",
            **self.DEFAULT_STYLE_KWARGS,
        )

    @property
    def axes(self):
        """List of all ax objects"""
        return [updater.ax for updater in self.plot_updaters]

    def display(self):
        top_widgets = widgets.HBox(
            children=[
                self.update_button,
                self.x_axis_style_widget,
            ]
        )
        display(top_widgets)
        display(self.invert_temp_axis, self.mode_widget)

        display(self.plots_tabs)

    def get_key(self, key):
        """Helper function which raises a more useful error message"""
        try:
            return self.app_data[key]
        except KeyError:
            raise KeyError(
                (
                    f"No key {key} available in app data. "
                    "Did you forget to load/run simulation first?"
                )
            )

    def get_data(self):
        mode = self.active_mc_mode
        if mode == "canonical":
            return self.get_key(self.KEYS.CANONICAL_MC_DATA)
        raise NotImplementedError(f"Mode not available: {mode}")

    def _on_x_axis_change(self, change):
        if utils.is_value_change(change):
            self._set_axis_type()

    @update_statusbar
    @utils.disable_cls_widget("update_button")
    def _on_update_click(self, b):
        with self.event_context(logger=logger):
            self._update_plot()

    def _update_plot(self):
        # do stuff
        with plt.style.context("seaborn"):
            for updater in self.plot_updaters:
                updater.update_plot()
            self._set_axis_type()

        for fig, out in zip(self.figures, self.plot_outputs):
            with out:
                clear_output(wait=True)
                display(fig)

    def _update_energy_variance_plot(self, ax):
        ax.clear()
        for run_i, data in enumerate(self.get_data(), start=1):
            x = np.array(data["temperature"])
            y = np.array(data["en_var"])
            ax.plot(x, y, "o--", label=f"Run {run_i}")

        if self.invert_temp_axis.value:
            ax.invert_xaxis()  # Temp goes from high->low

        ax.set_xlabel("Temperature (K)")
        ax.set_ylabel(r"$\langle E^2 \rangle - \langle E \rangle ^2$")
        ax.set_title("Energy Variance")
        self._post_process_ax(ax)

    def _update_heat_capacity_by_diff_plot(self, ax):
        ax.clear()
        for run_i, data in enumerate(self.get_data(), start=1):
            x = np.array(data["temperature"])
            en = _get_e_mean(data, normalize=False)
            filtered = savgol_filter(en, 5, 2)
            y = np.gradient(filtered) / np.gradient(x)
            ax.plot(x, y, "o--", label=f"Run {run_i}")

        if self.invert_temp_axis.value:
            ax.invert_xaxis()  # Temp goes from high->low

        ax.set_xlabel("Temperature (K)")
        ax.set_ylabel("Heat Capacity (eV/K)")
        ax.set_title("Heat Capacity from Avg. Energy Differentiation")
        self._post_process_ax(ax)

    def _update_heat_capacity_plot(self, ax):
        ax.clear()
        for run_i, data in enumerate(self.get_data(), start=1):
            x = np.array(data["temperature"])
            y = np.array(data["heat_capacity"])
            ax.plot(x, y, "o--", label=f"Run {run_i}")

        if self.invert_temp_axis.value:
            ax.invert_xaxis()  # Temp goes from high->low

        ax.set_xlabel("Temperature (K)")
        ax.set_ylabel("Heat Capacity (eV/K)")
        self._post_process_ax(ax)

    def _update_emin_plot(self, ax):
        ax.clear()
        # Get plot data
        for run_i, data in enumerate(self.get_data()):
            meta = data["meta"]
            natoms = meta["natoms"]
            x = np.array(data["temperature"])
            # Normalize energy to eV/atom
            y = np.array(data["emin"]) / natoms
            if run_i == 0:
                # Normalize to the first energy point of the first run
                E0 = y[0]
            y -= E0
            # Do the plot
            ax.plot(x, y, "o--", label=f"Run {run_i+1}")

        if self.invert_temp_axis.value:
            ax.invert_xaxis()  # Temp goes from high->low

        ax.set_xlabel("Temperature (K)")
        ax.set_ylabel(r"$E_{min} - E_0$ (eV/atom)")
        self._post_process_ax(ax)

    def _update_emean_plot(self, ax):
        ax.clear()
        # Get plot data
        for run_i, data in enumerate(self.get_data()):
            x = np.array(data["temperature"])
            y = _get_e_mean(data)
            if run_i == 0:
                # Normalize to the first energy point of the first run
                E0 = y[0]
            y -= E0
            # Do the plot
            ax.plot(x, y, "o--", label=f"Run {run_i+1}")
        if self.invert_temp_axis.value:
            ax.invert_xaxis()  # Temp goes from high->low
        ax.set_xlabel("Temperature (K)")
        ax.set_ylabel(r"$E_{mean} - E_0$ (eV/atom)")
        self._post_process_ax(ax)

    def _update_accept_rate_plot(self, ax):
        ax.clear()
        # Get plot data
        for run_i, data in enumerate(self.get_data()):
            x = np.array(data["temperature"])
            y = np.array(data["accept_rate"])

            # Do the plot
            ax.plot(x, y, "o--", label=f"Run {run_i+1}")

        if self.invert_temp_axis.value:
            ax.invert_xaxis()

        ax.set_xlabel("Temperature (K)")
        ax.set_ylabel("Accept rate")
        # Accept rate goes from [0, 1], so add a little extra
        # so the dots stay within the plot
        ax.set_ylim([-0.05, 1.05])
        self._post_process_ax(ax)

    @staticmethod
    def _post_process_ax(ax) -> None:
        """Post processing of the axis object, if the updater function choses."""
        legend = ax.legend(frameon=True)
        # This is a nicer legend color when using the seaborn dark grid
        legend.get_frame().set_facecolor("white")

    def _set_axis_type(self):
        value = self.x_axis_style_widget.value
        for ax in self.axes:
            ax.set_xscale(value)

    @property
    def active_mc_mode(self):
        return self.mode_widget.value

    def _make_savefig_box(self, fig):
        """Make an hbox with a save figure button and a text field
        to specify the figure name"""
        savefig_button = widgets.Button(description="Save Figure")
        fig_name_text = widgets.Text(description="Filename", value="figure.png")

        def _on_save_click(b):
            cwd = self.app_data[self.KEYS.CWD]
            fname = str(cwd / fig_name_text.value)

            fig.savefig(fname)
            logger.info("Saved figure to file: %s", fname)

        savefig_button.on_click(_on_save_click)
        # Hbox placing the button and text widget side-by-side
        box = widgets.HBox(children=[savefig_button, fig_name_text])
        return box


def _get_e_mean(data: dict, normalize=True) -> np.ndarray:
    meta = data["meta"]
    en = np.array(data["mean_energy"])
    if normalize:
        # Normalize energy to eV/atom
        natoms = meta["natoms"]
        en /= natoms
    return en



"""Module with some tools for timing stuff."""
from datetime import timedelta


def est_runtime_delta(prog, cur_runtime) -> timedelta:
    """Estimate the remaining runtime as a timedelta function.
    Rounds up to make things look nicer.

    Args:
        prog (float): The current progress, between 0 and 1.
        cur_runtime (float): How long the process has been running.
    """
    return timedelta(seconds=round(est_runtime(prog, cur_runtime)))


def est_runtime(prog: float, cur_runtime: float) -> float:
    """Estimate the remaining runtime.

    Args:
        prog (float): The current progress, between 0 and 1.
        cur_runtime (float): How long the process has been running.
    """
    # Example:
    # 30 % progress, 10 seconds => 10 / 30 = 1/3 s per %.
    # Est.: 1/3 * 70 % = 23.3 s
    if prog == 0:
        # cannot do an estimate with 0 progressed time.
        return 0.0

    return (cur_runtime / prog) * (1.0 - prog)



import logging
from collections import defaultdict
from typing import Dict, Any, Callable
from enum import Enum, IntEnum, auto
from pathlib import Path
import attr
from clease import jsonio
from clease_gui.logging_widget import register_logger

__all__ = ["save_app_data", "load_app_data", "AppDataKeys"]

logger = logging.getLogger(__name__)
register_logger(logger)


class AppDataKeys(str, Enum):
    """Collection of keys which (may) be in the app_data.
    Keys starting with an '_' will not be saved in the app state."""

    # "private" keys
    CWD = "_cwd"
    STATUS = "_status"
    DEV_MODE = "_dev_mode"
    STEP_OBSERVER = "_mc_step_obs"

    # Regular app data keys
    SUPERCELL = "supercell"
    SETTINGS = "settings"
    ECI = "eci"
    CANONICAL_MC_DATA = "canonical_mc_data"

    # The evaluator cannot be saved to file, so save it as private
    # instance of an Evaluate class
    EVALUATE = "_evaluator"

    @classmethod
    def is_key_private(cls, key: str) -> bool:
        """Check if a given key is considered 'private'"""
        return key.startswith("_")

    @classmethod
    def is_key_public(cls, key: str) -> bool:
        """Check if a given key is considered 'public'"""
        return not cls.is_key_private(key)

    @classmethod
    def iter_public_keys(cls):
        yield from filter(cls.is_key_public, cls)

    @classmethod
    def iter_private_keys(cls):
        yield from filter(cls.is_key_private, cls)


class Notifier(IntEnum):
    """Type of notification"""

    SET = auto()
    DELETE = auto()
    POP = auto()


@attr.define
class Notification:
    """A notification that something changed in the AppData"""

    key: str = attr.field()
    old_value: Any = attr.field()
    new_value: Any = attr.field()
    notifier: Notifier = attr.field()


class AppData(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Shortcut to AppDataKeys
        self.app_keys = AppDataKeys
        self._subscribers = defaultdict(list)

    def subscribe(self, key: str, func: Callable[[Notification], None]) -> None:
        self._subscribers[key].append(func)

    def notify_subscribers(self, key: str, change: Notification) -> None:
        for ii, func in enumerate(self._subscribers[key]):
            logger.debug("Notifying subscriber %d to key %s from %s", ii, key, change.notifier)
            func(change)

    def pop(self, key: str, **kwargs) -> Any:
        """Pop an item like in dict.pop(), and then notify anyone who subscribes to that key.
        Notificattion only occurs after the delete."""
        old_value = self.get(key, None)
        ret_val = super().pop(key, **kwargs)
        change = Notification(key, old_value, None, Notifier.POP)
        self.notify_subscribers(key, change)
        return ret_val

    def __setitem__(self, key: str, value: Any) -> None:
        """Change an item, and then notify anyone who subscribes to that key.
        Notificattion only occurs after the set."""
        old_value = self.get(key, None)
        ret_val = super().__setitem__(key, value)
        change = Notification(key, old_value, value, Notifier.SET)
        self.notify_subscribers(key, change)
        return ret_val

    def __delitem__(self, key: str) -> None:
        """Delete an item, and then notify anyone who subscribes to that key.
        Notificattion only occurs after the delete."""
        old_value = self.get(key, None)
        ret_val = super().__delitem__(key)
        change = Notification(key, old_value, None, Notifier.DELETE)
        self.notify_subscribers(key, change)
        return ret_val

    def update(self, *dicts, **kwargs):
        """Override the built-in update() dict method, to
        use the custom __setitem__ methods."""

        def _update(dct):
            for key, value in dct.items():
                self[key] = value

        for dct in dicts:
            _update(dct)
        _update(kwargs)


def save_app_data(app_data: Dict[str, Any], fname) -> None:
    fname = Path(fname)
    data = app_data.copy()
    to_remove = []
    for key in data:
        # Find keys which we don't want to save
        # Any keys starting with a "_" we say
        # we don't want to save
        if AppDataKeys.is_key_private(key):
            to_remove.append(key)
    for key in to_remove:
        data.pop(key, None)

    with fname.open("w") as file:
        jsonio.write_json(file, data)


def load_app_data(fname) -> AppData:
    fname = Path(fname)
    with fname.open() as file:
        return AppData(**jsonio.read_json(file))



from setuptools import setup, find_packages
from pathlib import Path


def read_requirements(filename):
    path = Path(__file__).parent / "requirements"
    with (path / filename).open("r") as f:
        return tuple(line.strip() for line in f if line.strip())


publish_requirements = read_requirements("publish_requirements.txt")
test_requirements = read_requirements("test_requirements.txt")
doc_requirements = read_requirements("doc_requirements.txt")
dev_requirements = read_requirements("dev_requirements.txt")

EXTRAS_REQUIRE = {
    "dev": dev_requirements,
    "publish": publish_requirements,
    "test": test_requirements,
    "doc": doc_requirements,
}

EXTRAS_REQUIRE["all"] = set(value for tup in EXTRAS_REQUIRE.values() for value in tup)

setup(
    packages=find_packages(include=["clease_gui", "clease_gui.*"]),
    extras_require=EXTRAS_REQUIRE,
)



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
import numpy as np
from clease.settings import ClusterExpansionSettings
from clease_gui import utils
import clease_gui as gui
from clease_gui.app_data import Notification
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar

from . import settings_buttons as swidgets

__all__ = ["CESettingsDashboard"]

logger = logging.getLogger(__name__)
gui.register_logger(logger)

DESC_WIDTH = "150px"
STYLE = {"description_width": DESC_WIDTH}


class CESettingsDashboard(BaseDashboard):
    def initialize(self) -> None:
        self.size_mode_out = widgets.Output()
        self.size_mode_widgets = {
            "fixed": swidgets.FixedSizeMode(output=self.size_mode_out),
            "supercell_factor": swidgets.SupercellFactorSizeMode(output=self.size_mode_out),
        }
        self.size_mode_b = widgets.Dropdown(
            options=[("Fixed", "fixed"), ("Supercell Factor", "supercell_factor")],
            value="supercell_factor",
            description="Size mode:",
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.size_mode_b.observe(self._on_size_mode_change)

        # Common settings, such as max cluster dia, background atoms and database name
        self.out_common = widgets.Output()
        self.common_settings_widget = swidgets.CommonSettingsButtons(
            self.app_data, output=self.out_common
        )
        self.common_settings_widget.display_widgets()

        self.update_settings_btn = utils.make_clickable_button(
            self._on_update_settings_click, description="Update Settings"
        )
        self.app_data.subscribe(self.KEYS.SETTINGS, self._enable_settings_button)

    def _enable_settings_button(self, change: Notification):
        if change.key == self.KEYS.SETTINGS:
            # Enable button depending if we have a settings
            is_disabled = change.new_value is None
            self.update_settings_btn.disabled = is_disabled

    @update_statusbar
    @utils.disable_cls_widget("update_settings_btn")
    def _on_update_settings_click(self, b):
        with self.event_context(logger=logger):
            self.update_settings()
            logger.info("Updated settings.")

    def update_settings(self) -> None:
        """Update the CE settings of the internal settings object (if it exists)."""
        # Fetch the live settings object (if it exists)
        settings = self.settings
        kwargs = self.get_settings_kwargs(updatable=True)
        for key, value in kwargs.items():
            setattr(settings, key, value)
        # We explicitly clear the cache, just for safety.
        # XXX: Should eventually be done in CLEASE itself,
        # it seems like this doesn't happen upon adjusting the MCD.
        settings.clear_cache()

    def display(self) -> None:
        super().display()
        display(self.out_common)

        # Buttons for size mode
        with self.size_mode_out:
            self.active_size_mode_widget.display_widgets()
        display(self.size_mode_b, self.size_mode_out)
        display(self.update_settings_btn)

    @property
    def active_size_mode(self):
        return self.size_mode_b.value

    @property
    def active_size_mode_widget(self):
        return self.size_mode_widgets[self.active_size_mode]

    def _on_size_mode_change(self, change):
        if utils.is_value_change(change):
            self._update_size_mode_widget()

    def _update_size_mode_widget(self):
        logger.debug("Updating active size mode widget")
        with self.size_mode_out:
            clear_output(wait=True)
            self.active_size_mode_widget.display_widgets()

    def get_settings_kwargs(self, updatable: bool = False):
        """Get the kwargs for settings. If updatable=True, only
        get the settings which may be updated."""
        size_kwargs = self.active_size_mode_widget.value
        common_settings_kwargs = self.common_settings_widget.value
        kwargs = dict(
            **common_settings_kwargs,
            **size_kwargs,
        )
        if updatable:
            # Remove any kwargs which may not be updated.
            # TODO: Lock this button, so it cannot be changed when a settings exists
            kwargs.pop("include_background_atoms", None)
        return kwargs

    def set_widgets_from_load(self, settings: ClusterExpansionSettings) -> None:
        self._load_settings(settings)

    def _load_settings(self, settings: ClusterExpansionSettings):
        # Guess the size mode
        if settings.size is None:
            logger.debug("Detected supercell factor mode")
            # use supercell factor
            self.size_mode_b.value = "supercell_factor"
            self._update_size_mode_widget()
            scf = settings.supercell_factor
            skew = settings.skew_threshold
            self.active_size_mode_widget.set_widget_value("supercell_factor", scf)
            self.active_size_mode_widget.set_widget_value("skew_threshold", skew)
        else:
            logger.debug("Detected fixed size mode")
            self.size_mode_b.value = "fixed"
            # self._update_size_mode_widget()
            logger.debug("Getting size value")
            value = settings.size  # in 3x3 matrix
            # Use diagonal
            value = np.diag(value)
            value = ", ".join(map(str, value))
            logger.debug("Setting widget %s to value %s", "size", value)
            self.active_size_mode_widget.set_widget_value("size", value)

        # Settings stored directly in the settings object
        all_keys = (
            "basis_func_type",
            "db_name",
            "max_cluster_size",
            "max_cluster_dia",
            "include_background_atoms",
        )
        for key in all_keys:
            value = get_value_for_widget(settings, key)
            logger.debug("Setting %s to %s", key, value)
            self.common_settings_widget.set_widget_value(key, value)


def get_value_for_widget(settings, key):
    """Get a value from the settings, and sanitize it for settings
    a widget state"""
    if key == "basis_func_type":
        return settings.basis_func_type.name
    if key in {
        "db_name",
        "max_cluster_size",
        "max_cluster_dia",
        "include_background_atoms",
    }:
        return getattr(settings, key)

    raise ValueError(f"Cannot deal with key: {key}")



import logging
from clease_gui import register_logger

base_logger = logging.getLogger(__name__)
register_logger(base_logger)

__all__ = ["log_error", "EventContextManager"]


def log_error(*args, logger=None, dev_mode=False, **kwargs) -> None:
    """Log an error, and determine logging level based on whether we are in dev mode or not.
    I.e., in dev_mode the traceback is included, but not in regular mode."""
    logger = logger or base_logger
    fnc = logger.exception if dev_mode else logger.error
    fnc(*args, **kwargs)


class EventContextManager:
    """Context manager for an event which shouldn't cause a crash of the GUI,
    such as when clicking a button, but an expected file is not found.
    The error will be logged, and the class can be inspected after the context
    to see if an error was found (and which type).

    Will catch any "Exception", but not "BaseException".

    Example:
    with EventContextManager() as cm:
        raise ValueError()
    assert cm.success is False
    assert isinstance(cm.exc_value, ValueError)
    """

    def __init__(self, dev_mode=False, logger=None):
        self.had_error = False
        # Record exception type and values
        self.exc_type = None
        self.exc_value = None

        self.is_done = False
        self._dev_mode = dev_mode
        self._logger = logger

    def __enter__(self):
        base_logger.debug("Entering event context.")
        return self

    def __exit__(self, exc_type, exc_value, trace) -> bool:
        self.exc_type = exc_type
        self.exc_value = exc_value
        if exc_type is not None:
            # We had an error. Log and suppress
            self.had_error = True
            log_error(exc_value, logger=self._logger, dev_mode=self._dev_mode)
        base_logger.debug("Leaving event context. Error detected? %s", self.had_error)
        self.is_done = True
        # Suppress any exception which is of type "Exception",
        # similar to except Exception: in a try/except block
        # i.e. we don't won't to block BaseExceptions such as KeyboardInterrupt or SystemExit
        return isinstance(exc_value, Exception)

    @property
    def success(self) -> bool:
        """Are we done, and completed with no error?"""
        return self.is_done and not self.had_error
